import subprocess

from je_editor.utils.exception.je_editor_exceptions import JEditorRunOnShellException


def run_on_shell(run_source):
    """
    :param run_source: string command to run
    :return: if error return result and True else return result and False
    """
    try:
        exec_result = subprocess.getoutput(run_source)
        return exec_result, False
    except Exception as error:
        return str(error), True



import sys


def except_hook(cls, exception, traceback):
    sys.__excepthook__(cls, exception, traceback)


class JEditorException(Exception):
    pass


class JEditorExecException(JEditorException):
    pass


class JEditorRunOnShellException(JEditorException):
    pass


class JEditorSaveFileException(JEditorException):
    pass


class JEditorOpenFileException(JEditorException):
    pass


class JEditorContentFileException(JEditorException):
    pass



import json
import os
from pathlib import Path
from threading import Lock

from je_editor.utils.exception.je_editor_exceptions import JEditorContentFileException

cwd = os.getcwd()
lock = Lock()

editor_data = {
    "last_file": None
}


def read_output_content():
    """
    read the editor content
    """
    try:
        lock.acquire()
        file_path = Path(cwd + "/je_editor_content.json")
        if file_path.exists() and file_path.is_file():
            with open(cwd + "/je_editor_content.json", "r+") as read_file:
                return read_file.read()
    except JEditorContentFileException:
        raise JEditorContentFileException
    finally:
        lock.release()


def write_output_content():
    """
    write the editor content
    """
    try:
        lock.acquire()
        with open(cwd + "/je_editor_content.json", "w+") as file_to_write:
            file_to_write.write(json.dumps(editor_data))
    except JEditorContentFileException:
        raise JEditorContentFileException
    finally:
        lock.release()


def save_content_and_quit(file):
    """
    set content data and write
    """
    editor_data["last_file"] = file
    write_output_content()


def open_content_and_start():
    """
    read data and set content
    """
    temp_content = read_output_content()
    if temp_content is not None:
        editor_data["last_file"] = json.loads(temp_content).get("last_file")
    return editor_data.get("last_file")



from je_editor.utils.file.open.open_file import read_file


def open_last_edit_file(file_from_output_content, code_editor):
    """
    :param file_from_output_content: readied file from output content
    :param code_editor the editor to insert file content
    :return readied file
    open last edit file
    if success open file
    insert file content to code_editor
    """
    temp_to_check_file = read_file(file_from_output_content)
    if temp_to_check_file is not None:
        code_editor.delete("1.0", "end-1c")
        code_editor.insert("end-1c", temp_to_check_file[1])
        return temp_to_check_file[0]



from je_editor.utils.file.save.save_file import SaveThread


def start_auto_save(auto_save_thread_object, current_file, code_editor):
    if auto_save_thread_object is not None:
        auto_save_thread_object.file = current_file
        return auto_save_thread_object
    elif current_file is not None and auto_save_thread_object is None:
        auto_save = SaveThread(current_file, code_editor)
        auto_save.start()
        return auto_save



import sys
import os.path
import queue
import shutil
import subprocess
from pathlib import Path
from threading import Thread
from tkinter import DISABLED
from tkinter import END
from tkinter import NORMAL

from je_editor.utils.exception.je_editor_exception_tag import file_not_fond_error
from je_editor.utils.exception.je_editor_exception_tag import python_not_found_error
from je_editor.utils.exception.je_editor_exceptions import JEditorExecException


class ExecManager(object):

    def __init__(self, run_result, process_error_function, main_window):
        self.main_window = main_window
        self.still_run_program = True
        self.run_result = run_result
        self.process_error_function = process_error_function
        self.process = None
        self.run_output_queue = queue.Queue()
        self.run_error_queue = queue.Queue()

    def exec_code(self, exec_file_name):
        """
        :param exec_file_name: string file will open to run
        :return: if error return result and True else return result and False
        """
        reformat_os_file_path = os.path.abspath(exec_file_name)
        try:
            if not Path(exec_file_name).exists():
                raise JEditorExecException(file_not_fond_error)
        except OSError as error:
            raise JEditorExecException(error)
        if sys.platform in ["linux", "linux2", "win32", "cygwin", "msys"]:
            python_path = shutil.which("python")
        else:
            python_path = shutil.which("python3")
        if python_path is None:
            raise JEditorExecException(python_not_found_error)
        exec_command = reformat_os_file_path
        self.process = subprocess.Popen([python_path, exec_command],
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        shell=False)
        read_program_output_from_thread = Thread(target=self.read_program_output_from_process)
        read_program_output_from_thread.setDaemon(True)
        read_program_output_from_thread.start()
        read_program_error_output_from_thread = Thread(target=self.read_program_error_output_from_process)
        read_program_error_output_from_thread.setDaemon(True)
        read_program_error_output_from_thread.start()
        print("execute: " + reformat_os_file_path)
        self.edit_tkinter_text()

    def edit_tkinter_text(self):
        self.run_result.configure(state=NORMAL)
        if not self.run_error_queue.empty():
            self.run_result.insert(END, self.run_error_queue.get())
        if not self.run_output_queue.empty():
            self.run_result.insert(END, self.run_output_queue.get())
        self.run_result.configure(state=DISABLED)
        if self.still_run_program:
            self.main_window.after(10, self.edit_tkinter_text)

    def exit_program(self):
        self.still_run_program = False

    def read_program_output_from_process(self):
        while self.still_run_program:
            program_output_data = self.process.stdout.raw.read(1024).decode()
            self.run_output_queue.put(program_output_data)

    def read_program_error_output_from_process(self):
        while self.still_run_program:
            program_error_output_data = self.process.stderr.raw.read(1024).decode()
            self.run_error_queue.put(program_error_output_data)



"""
default exception
"""
je_editor_error = "JEditor error"
"""
exec exception
"""
je_editor_exec_error = "JEditor exec error"
file_not_fond_error = "file not fond"
python_not_found_error = "python not found"
"""
shell exception
"""
je_editor_shell_error = "JEditor run on shell error"
"""
file exception
"""
je_editor_save_file_error = "JEditor save file error"
je_editor_open_file_error = "JEditor open file error"
je_editor_content_file_error = "JEditor content file error"



from je_editor.utils.editor_content.content_save import save_content_and_quit


def close_event(check_current_file, tkinter_window, exec_manager):
    """
    :param exec_manager:
    :param check_current_file: check have current file? if have; save and quit else only quit
    :param tkinter_window: the tkinter window will close
    :return: no return value
    """
    if check_current_file is not None:
        save_content_and_quit(check_current_file)
    exec_manager.exit_program()
    tkinter_window.destroy()



import os
import time
from pathlib import Path
from threading import Lock
from threading import Thread
from tkinter import filedialog

from je_editor.utils.exception.je_editor_exceptions import JEditorSaveFileException

cwd = os.getcwd()
lock = Lock()


def write_file(file, content):
    """
    :param file: file we want to write
    :param content: content write in file
    try
        lock thread
        if file not empty string
            write content to file
    finally
        release lock
    """
    content = str(content)
    try:
        lock.acquire()
        if file != "":
            with open(file, "w+") as file_to_write:
                file_to_write.write(content)
    except JEditorSaveFileException:
        raise JEditorSaveFileException
    finally:
        lock.release()


def save_file(content):
    """
    :param content: content we want to write or ""
    :return: choose file
    open tkinter ask save file dialog
    if not choose
        len(file) = 0 or ""
        :return ""
    """
    file = filedialog.asksaveasfilename(title="Save File",
                                        initialdir=cwd,
                                        defaultextension="*.*",
                                        filetypes=(("all files", "*.*"), ("je editor files", "*.jee")))
    if len(file) == 0:
        return ""
    write_file(file, content)
    return file


class SaveThread(Thread):

    def __init__(self, file, tkinter_text, auto_save=False):
        """
        :param file: file we want to auto save
        :param tkinter_text: tkinter text
        :param auto_save: not need to change
        """
        super().__init__()
        self.file = file
        self.path = None
        self.tkinter_text = tkinter_text
        self.auto_save = auto_save
        # set daemon
        self.setDaemon(True)

    def run(self):
        # loop and save current edit file
        print("auto save start")
        if self.file is not None:
            self.auto_save = True
            self.path = Path(self.file)
        while self.auto_save:
            time.sleep(15)
            if self.path.exists() and self.path.is_file():
                print("auto saved file: ", str(self.file))
                write_file(self.file, self.tkinter_text.get("1.0", "end-1c"))
            else:
                break



from je_editor.utils.file.save.save_file import save_file


def save_file_to_open(code_editor):
    """
    :return saved file
    show save file dialog
    if saved
        change current file to new file
        start auto save
    """
    temp_to_check_file = save_file(code_editor.get("1.0", "end-1c"))
    if temp_to_check_file is not None and temp_to_check_file != "":
        return temp_to_check_file



import tkinter
from tkinter import E
from tkinter import Menu
from tkinter import N
from tkinter import S
from tkinter import Text
from tkinter import Tk
from tkinter import W
from tkinter import ttk

from je_editor.ui.ui_event.auto_save.start_auto_save.start_auto_save import start_auto_save
from je_editor.ui.ui_event.close.close_event import close_event
from je_editor.ui.ui_event.execute.execute_code.exec_code import execute_code
from je_editor.ui.ui_event.execute.execute_shell_command.run_on_shell import execute_shell_command
from je_editor.ui.ui_event.open_file.open_file_to_read.open_file_to_read import open_file_to_read
from je_editor.ui.ui_event.open_file.open_last_edit_file.open_last_edit_file import open_last_edit_file
from je_editor.ui.ui_event.save_file.save_file_to_open.save_file_to_open import save_file_to_open
from je_editor.utils.code_tag.tag_keyword import HighlightText
from je_editor.utils.editor_content.content_save import open_content_and_start
from je_editor.utils.text_process.program_exec.exec_text import ExecManager
from je_editor.utils.text_process.program_exec.process_error import process_error_text


def start_editor(use_theme=None):
    EditorMain(use_theme=use_theme).start_editor()


class EditorMain(object):

    # start auto save
    def start_auto_save(self):
        self.auto_save = start_auto_save(self.auto_save, self.current_file, self.code_editor)

    # start editor and start auto save if auto save not start
    def start_editor(self):
        self.start_auto_save()
        self.main_window.mainloop()

    # editor close event
    def close_event(self):
        close_event(self.current_file, self.main_window, self.exec_manager)

    # editor open file
    def open_file_to_read(self, event=None):
        temp = open_file_to_read(self.code_editor)
        self.file_from_output_content = temp
        self.current_file = temp
        self.highlight_text.search()
        self.start_auto_save()

    # save editor file
    def save_file_to_open(self, event=None):
        self.current_file = save_file_to_open(self.code_editor)
        self.start_auto_save()

    def open_last_edit_file(self):
        self.highlight_text.search()
        return open_last_edit_file(self.file_from_output_content, self.code_editor)

    def exec_code(self, event=None):
        execute_code(self.run_result, self.current_file, self.save_file_to_open, self.exec_manager)

    def run_on_shell(self, event=None):
        execute_shell_command(self.run_result, self.code_editor)

    def show_popup_menu(self, event):
        """
        :param event: tkinter event bind Button-3
        try
            show popup menu
        finally
            release popup menu
        """
        try:
            self.popup_menu.tk_popup(event.x_root, event.y_root)
        finally:
            self.popup_menu.grab_release()

    # default event
    def do_test(self, event=None):
        self.test_run = True
        print("test")

    def __init__(self, use_theme=None, main_window=Tk()):
        """
        :param main_window: Tk instance
        """
        # style
        self.style = ttk.Style()
        if use_theme is not None:
            self.style.theme_use(use_theme)
        # set main window title and add main frame
        self.main_window = main_window
        self.main_window.title("je_editor")
        self.code_edit_frame = ttk.Frame(self.main_window, padding="3 3 12 12")
        self.code_edit_frame.grid(column=0, row=0, sticky=(N, W, E, S))
        self.run_result_frame = ttk.Frame(self.main_window, padding="3 3 12 12")
        self.run_result_frame.grid(column=0, row=1, sticky=(N, W, E, S))
        # Text start and end position
        self.start_position = "1.0"
        self.end_position = "end-1c"
        # set code edit
        self.code_editor = Text(self.code_edit_frame, undo=True, autoseparators=True, maxundo=-1)
        self.code_editor_scrollbar_y = ttk.Scrollbar(orient="vertical", command=self.code_editor.yview)
        self.code_editor["yscrollcommand"] = self.code_editor_scrollbar_y.set
        self.code_editor.grid(column=0, row=0, sticky=(N, W, E, S))
        self.code_editor_scrollbar_y.grid(column=1, row=0)
        self.code_editor.configure(state="normal")
        # run result
        self.run_result = Text(self.run_result_frame)
        self.run_result_scrollbar_y = ttk.Scrollbar(orient="vertical", command=self.run_result.yview)
        self.run_result["yscrollcommand"] = self.run_result_scrollbar_y.set
        self.run_result.grid(column=0, row=1, sticky=(N, W, E, S))
        self.run_result_scrollbar_y.grid(column=1, row=1)
        self.run_result.configure(state="disabled")
        self.run_result.bind("<1>", lambda event: self.run_result.focus_set())
        self.exec_manager = ExecManager(run_result=self.run_result,
                                        process_error_function=process_error_text,
                                        main_window=self.main_window)
        # Menubar
        # Main menu
        self.menu = tkinter.Menu(self.main_window)
        # File menu
        self.file_menu = tkinter.Menu(self.menu, tearoff=0)
        self.file_menu.add_command(label="Save File", command=self.save_file_to_open)
        self.file_menu.add_command(label="Open File", command=self.open_file_to_read)
        # Function menu
        self.function_menu = tkinter.Menu(self.menu, tearoff=0)
        self.function_menu.add_command(label="Run", command=self.exec_code)
        self.function_menu.add_command(label="Run on shell", command=self.run_on_shell)
        # Popup menu
        self.popup_menu = Menu(self.main_window, tearoff=0)
        self.popup_menu.add_command(label="Run", command=self.exec_code)
        self.popup_menu.add_command(label="Run on shell", command=self.run_on_shell)
        self.popup_menu.add_separator()
        self.popup_menu.add_command(label="Save File", command=self.save_file_to_open)
        self.popup_menu.add_command(label="Open File", command=self.open_file_to_read)
        self.main_window.bind("<Button-3>", self.show_popup_menu)
        # add and config
        self.menu.add_cascade(label="File", menu=self.file_menu)
        self.menu.add_cascade(label="Function", menu=self.function_menu)
        self.main_window.config(menu=self.menu)
        # set resize
        self.code_edit_frame.columnconfigure(0, weight=1)
        self.code_edit_frame.rowconfigure(0, weight=1)
        self.run_result_frame.columnconfigure(0, weight=1)
        self.run_result_frame.rowconfigure(1, weight=1)
        self.main_window.columnconfigure(0, weight=1)
        self.main_window.rowconfigure(0, weight=1)
        # Highlight word
        self.highlight_text = HighlightText(self.code_editor)
        # current file
        self.current_file = None
        # file to output content
        self.file_from_output_content = open_content_and_start()
        if self.file_from_output_content is not None:
            self.current_file = self.open_last_edit_file()
            self.highlight_text.search()
        # close event
        self.main_window.protocol("WM_DELETE_WINDOW", self.close_event)
        # bind
        self.main_window.bind("<Control-Key-o>", self.open_file_to_read)
        self.main_window.bind("<Control-Key-s>", self.save_file_to_open)
        self.main_window.bind("<Control-Key-F5>", self.exec_code)
        self.main_window.bind("<Control-Key-F6>", self.run_on_shell)
        # is this test run?
        self.test_run = False
        # Auto save thread
        self.auto_save = None
        if self.current_file is not None:
            self.auto_save = start_auto_save(self.auto_save, self.current_file, self.code_editor)





from je_editor.utils.exception.je_editor_exceptions import *
# exec code
from je_editor.utils.text_process.program_exec.exec_text import ExecManager
from je_editor.utils.text_process.shell.shell_text import run_on_shell
# editor content
from je_editor.utils.editor_content.content_save import write_output_content
from je_editor.utils.editor_content.content_save import read_output_content
# file
from je_editor.utils.file.open.open_file import open_file
from je_editor.utils.file.save.save_file import save_file
# start editor
from je_editor.ui.editor_main_ui.tkinter_editor import start_editor
